"""Test the remote operation response object."""

import json
import unittest
from uuid import UUID

from mitsubishi_connect_client.remote_operation_response import RemoteOperationResponse

from . import sample_remote_operaton_response


class TestRemoteOperationResponse(unittest.TestCase):
    """Test the remote operation response object."""

    def test_from_text(self) -> None:
        """Test the from_text method."""
        loaded = json.loads(json.dumps(sample_remote_operaton_response))
        response = RemoteOperationResponse(**loaded)
        self.assertEqual(
            response.event_id, UUID("59668d8a-6426-4691-b61b-3c87d206d3f9")
        )
        self.assertEqual(response.status_timestamp, "2024-03-14T12:34:56.789Z")
        self.assertEqual(response.start_time, "2024-03-14T12:34:56.789Z")
        self.assertEqual(response.operation_type, "engineOff")
        self.assertEqual(response.vin, "1234567890ABCDEFG")
        self.assertEqual(response.state, "1")
        self.assertEqual(response.status, "success")



"""File for tests."""

sample_vehicle = {
    "vehicles": [
        {
            "vin": "vin",
            "dateOfSale": "2024-12-29",
            "primaryUser": True,
            "make": "make",
            "model": "model",
            "year": "2024",
            "exteriorColorCode": "exteriorColorCode",
            "exteriorColor": "exteriorColor",
            "simState": "simState",
            "modelDescription": "modelDescription",
            "country": "country",
            "region": "region",
            "alphaThreeCountryCode": "alphaThreeCountryCode",
            "countryName": "countryName",
            "isFleet": False,
        }
    ]
}
sample_vehicle_state = {
    "vin": "1234567890ABCDEFG",
    "ts": "2024-03-14T12:34:56.789Z",
    "state": {
        "extLocMap": {"lon": 123.456, "lat": 456.789, "ts": "1678886400000"},
        "cst": "1",
        "tuState": "1",
        "ods": "0",
        "ignitionState": "0",
        "odo": [{"2025-02-09 15:14:49": "1223"}, {"2025-02-10 20:54:33": "1242"}],
        "theftAlarm": "OFF",
        "svla": "0",
        "svtb": "0",
        "diagnostic": "0",
        "privacy": "0",
        "temp": "1",
        "factoryReset": "0",
        "tuStateTS": "1739934731691",
        "ignitionStateTs": "1739913349354",
        "timezone": "UTC",
        "accessible": True,
        "chargingControl": {
            "cruisingRangeCombined": "200",
            "eventTimestamp": "1678886400000",
        },
    },
}

sample_remote_operaton_response = {
    "eventId": "59668d8a-6426-4691-b61b-3c87d206d3f9",
    "statusTimestamp": "2024-03-14T12:34:56.789Z",
    "startTime": "2024-03-14T12:34:56.789Z",
    "operationType": "engineOff",
    "vin": "1234567890ABCDEFG",
    "state": "1",
    "status": "success",
}


sample_vehicle_status = {
    "vhr": [
        {
            "cid": "a7ddec3e-575f-4ea4-9038-c1692caecac2",
            "vin": "1234567890ABCDEFG",
            "operation": "vehicleStatus",
            "ts": 1739934753,
            "dt": {
                "diagnostic": {
                    "breakWarn": {
                        "name": "Brake Warning",
                        "displayMessage": "break_good",
                        "value": "0",
                        "warning": False,
                    },
                    "odos": {
                        "name": "Mileage Unit",
                        "displayMessage": "odos_miles",
                        "value": "0",
                    },
                    "engineOilWarn": {
                        "name": "Engine Oil Pressure",
                        "displayMessage": "engineOil_good",
                        "value": "0",
                        "warning": False,
                    },
                    "digsts": "1739934741242",
                    "tireStatus": {
                        "name": "TPMS Warning",
                        "displayMessage": "tirepressure_adequately",
                        "value": "0",
                        "warning": False,
                    },
                    "availRange": {
                        "name": "Available Range",
                        "displayMessage": "availRange_miles",
                        "value": "--",
                    },
                    "odo": {
                        "name": "Mileage",
                        "displayMessage": "odo_miles",
                        "value": "1447",
                    },
                    "spd": {
                        "name": "Vehicle Speed",
                        "displayMessage": "vehicle_speed",
                        "value": "0",
                    },
                    "igst": {
                        "name": "IG Status",
                        "displayMessage": "igst_off",
                        "value": "0",
                    },
                },
                "message": "SNAPSHOT",
                "vehicleStatus": {
                    "doorStatus": {
                        "id": "doorStatus",
                        "name": "Door Status",
                        "doors": [
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorFrontLeft",
                                    "value": "0",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorHood",
                                    "value": "1",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorFrontRight",
                                    "value": "2",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorRearLeft",
                                    "value": "3",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorRearRight",
                                    "value": "4",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                            {
                                "position": {
                                    "name": "Door Position",
                                    "displayMessage": "doorTrunk",
                                    "value": "5",
                                },
                                "state": {
                                    "name": "Door State",
                                    "displayMessage": "#msg_close_close",
                                    "value": "0",
                                },
                            },
                        ],
                    },
                    "lightStatus": {
                        "id": "lightStatus",
                        "name": "Lights Status",
                        "lights": [
                            {
                                "state": {
                                    "name": "Light State",
                                    "displayMessage": "#msg_off_off",
                                    "value": "3",
                                },
                                "position": {
                                    "name": "Light Position",
                                    "displayMessage": "lightHead",
                                    "value": "0",
                                },
                            }
                        ],
                    },
                },
            },
        }
    ]
}



"""Define a vehicle object."""

from __future__ import annotations

from datetime import date  # noqa: TC003 needed for pydantic

from pydantic import BaseModel, Field


class Vehicle(BaseModel):
    """Define a vehicle object."""

    vin: str
    date_of_sale: date = Field(..., alias="dateOfSale")
    primary_user: bool = Field(..., alias="primaryUser")
    make: str
    model: str
    year: int
    exterior_color_code: str = Field(..., alias="exteriorColorCode")
    exterior_color: str = Field(..., alias="exteriorColor")
    sim_state: str = Field(..., alias="simState")
    model_description: str = Field(..., alias="modelDescription")
    country: str
    region: str
    alpha_three_country_code: str = Field(..., alias="alphaThreeCountryCode")
    country_name: str = Field(..., alias="countryName")
    is_fleet: bool = Field(..., alias="isFleet")


class VechiclesResponse(BaseModel):
    """Define a vehicles response object."""

    vehicles: list[Vehicle]



"""Test the mitsubishi connect client."""

import json
import unittest
from unittest.mock import AsyncMock, MagicMock, patch

import pytest

from mitsubishi_connect_client.mitsubishi_connect_client import MitsubishiConnectClient
from mitsubishi_connect_client.token_state import TokenState

from . import (
    sample_remote_operaton_response,
    sample_vehicle,
    sample_vehicle_state,
    sample_vehicle_status,
)


class TestMitsubishiConnectClient(unittest.IsolatedAsyncioTestCase):
    """Test the mitsubishi connect client."""

    async def asyncSetUp(self) -> None:
        """Set up the test."""
        _client = MitsubishiConnectClient("username", "password")
        _token_dict = {
            "access_token": "12345",
            "refresh_token": "54321",
            "expires_in": 3600,
            "token_type": "bearer",
            "refresh_expires_in": 3600,
            "accountDN": "1256",
        }
        _token = TokenState(**_token_dict)
        self._token_dict = _token_dict
        self._client = _client
        self._token = _token

    @patch("aiohttp.ClientSession.request")
    async def test_login(self, mock_post: MagicMock) -> None:
        """Test the login function."""
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(self._token_dict)
        mock_post.return_value.__aenter__.return_value = mock_response
        await self._client.login()
        assert self._client.token == self._token

    @patch("aiohttp.ClientSession.request")
    async def test_refresh_token(self, mock_post: MagicMock) -> None:
        """Test the refresh_token function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(self._token_dict)
        mock_post.return_value.__aenter__.return_value = mock_response
        await self._client.refresh_token()
        assert self._client.token == self._token

    @patch("aiohttp.ClientSession.request")
    async def test_get_vehicles(self, mock_get: MagicMock) -> None:
        """Test the get_vehicles function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_vehicle)
        mock_get.return_value.__aenter__.return_value = mock_response
        vehicles = await self._client.get_vehicles()
        assert vehicles.vehicles[0].vin == "vin"

    @patch("aiohttp.ClientSession.request")
    async def test_get_vehicle_state(self, mock_get: MagicMock) -> None:
        """Test the get_vehicle_state function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_vehicle_state)
        mock_get.return_value.__aenter__.return_value = mock_response
        vehicle_state = await self._client.get_vehicle_state("test_vin")
        assert vehicle_state.vin == "1234567890ABCDEFG"

    @patch("aiohttp.ClientSession.request")
    async def test_stop_engine(self, mock_post: MagicMock) -> None:
        """Test the stop_engine function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_remote_operaton_response)
        mock_post.return_value.__aenter__.return_value = mock_response
        response = await self._client.stop_engine("test_vin")
        assert response.status == "success"

    @patch("aiohttp.ClientSession.request")
    async def test_flash_lights(self, mock_post: MagicMock) -> None:
        """Test the flash_lights function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_remote_operaton_response)
        mock_post.return_value.__aenter__.return_value = mock_response
        response = await self._client.flash_lights("test_vin")
        assert response.status == "success"

    @patch("aiohttp.ClientSession.request")
    async def test_start_engine(self, mock_post: MagicMock) -> None:
        """Test the start_engine function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_remote_operaton_response)
        mock_post.return_value.__aenter__.return_value = mock_response
        response = await self._client.start_engine("test_vin")
        assert response.status == "success"

    @patch("aiohttp.ClientSession.request")
    async def test_unlock_vehicle(self, mock_post: MagicMock) -> None:
        """Test the unlock_vehicle function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_remote_operaton_response)
        mock_post.return_value.__aenter__.return_value = mock_response
        response = await self._client.unlock_vehicle("test_vin", "test_pin_token")
        assert response.status == "success"

    @patch("aiohttp.ClientSession.request")
    async def test_get_nonce(self, mock_post: MagicMock) -> None:
        """Test the get_nonce function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = '{"serverNonce": "test_server_nonce"}'
        mock_post.return_value.__aenter__.return_value = mock_response
        nonce = await self._client.get_nonce("test_vin")
        assert "clientNonce" in nonce
        assert nonce["serverNonce"] == "test_server_nonce"

    @patch("aiohttp.ClientSession.request")
    async def test_get_pin_token(self, mock_post: MagicMock) -> None:
        """Test the get_pin_token function."""
        self._client.token = self._token

        # Mock the get_nonce function
        mock_get_nonce = AsyncMock()
        mock_get_nonce.return_value = {
            "clientNonce": "test_client_nonce",
            "serverNonce": "test_server_nonce",
        }
        self._client.get_nonce = mock_get_nonce

        # Mock the API response for getting the PIN token
        mock_response = AsyncMock()
        mock_response.text.return_value = '{"pinToken": "test_pin_token"}'
        mock_post.return_value.__aenter__.return_value = mock_response

        pin_token = await self._client.get_pin_token("test_vin", "test_pin")
        assert pin_token == "test_pin_token"  # noqa: S105

    @patch("aiohttp.ClientSession.request")
    async def test_get_status(self, mock_get: MagicMock) -> None:
        """Test the get_status function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_vehicle_status)
        mock_get.return_value.__aenter__.return_value = mock_response
        vehicle_status = await self._client.get_status("test_vin")
        assert vehicle_status is not None

    @patch("aiohttp.ClientSession.request")
    async def test_bad_response(self, mock_get: MagicMock) -> None:
        """Test a bad response function."""
        self._client.token = self._token
        mock_response = AsyncMock()
        mock_response.text.return_value = json.dumps(sample_vehicle_status)
        mock_get.return_value.__aenter__.return_value = mock_response
        mock_response.ok = False
        mock_response.status = 400
        with pytest.raises(Exception):  # noqa: B017, PT011
            await self._client.get_status("test_vin")

    def test_generate_client_nonce_base64(self) -> None:
        """Test the generate_client_nonce_base64 function."""
        nonce = self._client._generate_client_nonce_base64()
        assert isinstance(nonce, str)
        forty_four = 44
        assert len(nonce) == forty_four

    def test_generate_hash(self) -> None:
        """Test the generate_hash function."""
        client_nonce = self._client._generate_client_nonce_base64()
        server_nonce = self._client._generate_client_nonce_base64()
        pin = "1234"
        hash_value = self._client._generate_hash(client_nonce, server_nonce, pin)
        assert isinstance(hash_value, str)



"""Run a test against the live api."""

import os

import pytest

from mitsubishi_connect_client.mitsubishi_connect_client import MitsubishiConnectClient
from mitsubishi_connect_client.remote_operation_response import RemoteOperationResponse
from mitsubishi_connect_client.vehicle_state import VehicleState


@pytest.mark.skip(reason="Live testing disabled.")
async def test_async() -> None:
    """Run a test against the live api."""
    _username = os.environ["MITSUBISHI_USERNAME"]
    _password = os.environ["MITSUBISHI_PASSWORD"]
    _client = MitsubishiConnectClient(_username, _password)

    pin = os.environ["MITSUBISHI_PIN"]
    await _client.login()
    assert _client.token is not None

    vehicles_response = await _client.get_vehicles()
    assert len(vehicles_response.vehicles) > 0

    vehicle_state = await _client.get_vehicle_state(vehicles_response.vehicles[0].vin)
    assert isinstance(vehicle_state, VehicleState)

    pin_token = await _client.get_pin_token(vehicles_response.vehicles[0].vin, pin)
    assert isinstance(pin_token, str)

    response = await _client.unlock_vehicle(
        vehicles_response.vehicles[0].vin, pin_token
    )
    assert isinstance(response, RemoteOperationResponse)



"""Test the vehicle state object."""

import unittest
from datetime import UTC, datetime

from mitsubishi_connect_client.vehicle_state import (
    ChargingControl,
    ExtLocMap,
    State,
    VehicleState,
)


class TestVehicleState(unittest.TestCase):
    """Test the vehicle state object."""

    def test_ext_loc_map(self) -> None:
        """Test the ExtLocMap model."""
        data = {"lon": 123.456, "lat": 456.789, "ts": "2023-03-15T13:20:00Z"}
        ext_loc_map = ExtLocMap(**data)
        self.assertEqual(ext_loc_map.lon, 123.456)
        self.assertEqual(ext_loc_map.lat, 456.789)
        self.assertEqual(ext_loc_map.ts, datetime(2023, 3, 15, 13, 20, tzinfo=UTC))

    def test_charging_control(self) -> None:
        """Test the ChargingControl model."""
        data = {
            "cruisingRangeCombined": "200",
            "eventTimestamp": "2023-03-15T13:20:00Z",
        }
        charging_control = ChargingControl(**data)  # type: ignore This is a unit test.
        self.assertEqual(charging_control.cruising_range_combined, "200")
        self.assertEqual(
            charging_control.event_timestamp, datetime(2023, 3, 15, 13, 20, tzinfo=UTC)
        )

    def test_state(self) -> None:
        """Test the State model."""
        ext_loc_map_data = {
            "lon": 123.456,
            "lat": 456.789,
            "ts": "2023-03-15T13:20:00Z",
        }
        charging_control_data = {
            "cruisingRangeCombined": "200",
            "eventTimestamp": "2023-03-15T13:20:00Z",
        }
        data = {
            "extLocMap": ext_loc_map_data,
            "cst": "1",
            "tuState": "1",
            "ods": "0",
            "ignitionState": "0",
            "odo": [{"2025-02-09 15:14:49": "1223"}, {"2025-02-10 20:54:33": "1242"}],
            "theftAlarm": "OFF",
            "svla": "0",
            "svtb": "0",
            "diagnostic": "0",
            "privacy": "0",
            "temp": "1",
            "factoryReset": "0",
            "tuStateTS": "2025-02-22T15:12:11.691Z",
            "ignitionStateTs": "2025-02-22T09:15:49.354Z",
            "timezone": "UTC",
            "accessible": True,
            "chargingControl": charging_control_data,
        }

        state = State(**data)
        self.assertEqual(state.ext_loc_map.lon, 123.456)
        self.assertEqual(state.ext_loc_map.lat, 456.789)
        self.assertEqual(
            state.ext_loc_map.ts, datetime(2023, 3, 15, 13, 20, tzinfo=UTC)
        )
        self.assertEqual(state.cst, True)
        self.assertEqual(state.tu_state, True)
        self.assertEqual(state.ods, False)
        self.assertEqual(state.ignition_state, False)
        self.assertEqual(
            state.odo,
            [
                {datetime(2025, 2, 9, 15, 14, 49): 1223},  # noqa: DTZ001
                {datetime(2025, 2, 10, 20, 54, 33): 1242},  # noqa: DTZ001
            ],
        )
        self.assertEqual(state.theft_alarm, False)
        self.assertEqual(state.svla, False)
        self.assertEqual(state.svtb, False)
        self.assertEqual(state.diagnostic, False)
        self.assertEqual(state.privacy, False)
        self.assertEqual(state.temp, "1")
        self.assertEqual(state.factory_reset, False)
        self.assertEqual(
            state.tu_state_ts, datetime(2025, 2, 22, 15, 12, 11, 691000, tzinfo=UTC)
        )
        self.assertEqual(
            state.ignition_state_ts,
            datetime(2025, 2, 22, 9, 15, 49, 354000, tzinfo=UTC),
        )
        self.assertEqual(state.timezone, "UTC")
        self.assertEqual(state.accessible, True)
        self.assertEqual(state.charging_control.cruising_range_combined, "200")
        self.assertEqual(
            state.charging_control.event_timestamp,
            datetime(2023, 3, 15, 13, 20, tzinfo=UTC),
        )

    def test_state_with_none_ignition_state_ts(self) -> None:
        """Test the State model when ignition_state_ts is none."""
        ext_loc_map_data = {
            "lon": 123.456,
            "lat": 456.789,
            "ts": "2023-03-15T13:20:00Z",
        }
        charging_control_data = {
            "cruisingRangeCombined": "200",
            "eventTimestamp": "2023-03-15T13:20:00Z",
        }
        data = {
            "extLocMap": ext_loc_map_data,
            "cst": "1",
            "tuState": "1",
            "ods": "0",
            "ignitionState": "0",
            "odo": [{"2025-02-09 15:14:49": "1223"}, {"2025-02-10 20:54:33": "1242"}],
            "theftAlarm": "OFF",
            "svla": "0",
            "svtb": "0",
            "diagnostic": "0",
            "privacy": "0",
            "temp": "1",
            "factoryReset": "0",
            "tuStateTS": "2025-02-22T15:12:11.691Z",
            "ignitionStateTs": None,
            "timezone": "UTC",
            "accessible": True,
            "chargingControl": charging_control_data,
        }
        state = State(**data)
        self.assertEqual(state.ignition_state_ts, None)

    def test_vehicle_state(self) -> None:
        """Test the VehicleState model."""
        ext_loc_map_data = {
            "lon": 123.456,
            "lat": 456.789,
            "ts": "2023-03-15T13:20:00Z",
        }
        charging_control_data = {
            "cruisingRangeCombined": "200",
            "eventTimestamp": "2023-03-15T13:20:00Z",
        }
        state_data = {
            "extLocMap": ext_loc_map_data,
            "cst": "1",
            "tuState": "1",
            "ods": "0",
            "ignitionState": "0",
            "odo": [{"2025-02-09 15:14:49": "1223"}, {"2025-02-10 20:54:33": "1242"}],
            "theftAlarm": "OFF",
            "svla": "0",
            "svtb": "0",
            "diagnostic": "0",
            "privacy": "0",
            "temp": "1",
            "factoryReset": "0",
            "tuStateTS": "2025-02-22T15:12:11.691Z",
            "ignitionStateTs": "2025-02-22T09:15:49.354Z",
            "timezone": "UTC",
            "accessible": True,
            "chargingControl": charging_control_data,
        }
        data = {
            "vin": "1234567890ABCDEFG",
            "ts": "2024-03-14T12:34:56.789Z",
            "state": state_data,
        }
        vehicle_state = VehicleState(**data)
        self.assertEqual(vehicle_state.vin, "1234567890ABCDEFG")
        self.assertEqual(
            vehicle_state.ts, datetime(2024, 3, 14, 12, 34, 56, 789000, tzinfo=UTC)
        )
        self.assertEqual(vehicle_state.state.ext_loc_map.lon, 123.456)
        self.assertEqual(vehicle_state.state.ext_loc_map.lat, 456.789)
        self.assertEqual(
            vehicle_state.state.ext_loc_map.ts,
            datetime(2023, 3, 15, 13, 20, tzinfo=UTC),
        )
        self.assertEqual(vehicle_state.state.cst, True)
        self.assertEqual(vehicle_state.state.tu_state, True)
        self.assertEqual(vehicle_state.state.ods, False)
        self.assertEqual(vehicle_state.state.ignition_state, False)
        self.assertEqual(
            vehicle_state.state.odo,
            [
                {datetime(2025, 2, 9, 15, 14, 49): 1223},  # noqa: DTZ001
                {datetime(2025, 2, 10, 20, 54, 33): 1242},  # noqa: DTZ001
            ],
        )
        self.assertEqual(vehicle_state.state.theft_alarm, False)
        self.assertEqual(vehicle_state.state.svla, False)
        self.assertEqual(vehicle_state.state.svtb, False)
        self.assertEqual(vehicle_state.state.diagnostic, False)
        self.assertEqual(vehicle_state.state.privacy, False)
        self.assertEqual(vehicle_state.state.temp, "1")
        self.assertEqual(vehicle_state.state.factory_reset, False)
        self.assertEqual(
            vehicle_state.state.tu_state_ts,
            datetime(2025, 2, 22, 15, 12, 11, 691000, tzinfo=UTC),
        )
        self.assertEqual(
            vehicle_state.state.ignition_state_ts,
            datetime(2025, 2, 22, 9, 15, 49, 354000, tzinfo=UTC),
        )
        self.assertEqual(vehicle_state.state.timezone, "UTC")
        self.assertEqual(vehicle_state.state.accessible, True)
        self.assertEqual(
            vehicle_state.state.charging_control.cruising_range_combined, "200"
        )
        self.assertEqual(
            vehicle_state.state.charging_control.event_timestamp,
            datetime(2023, 3, 15, 13, 20, tzinfo=UTC),
        )



"""Define a vehicle status object."""

from __future__ import annotations

from pydantic import BaseModel, Field


class BreakWarn(BaseModel):
    """Define a break warn object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str
    warning: bool


class Odos(BaseModel):
    """Define an odo object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class EngineOilWarn(BaseModel):
    """Define an engine oil warn object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str
    warning: bool


class TireStatus(BaseModel):
    """Define a tire status object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str
    warning: bool


class AvailRange(BaseModel):
    """Define an avail range object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Odo(BaseModel):
    """Define an odo object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Spd(BaseModel):
    """Define a spd object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Igst(BaseModel):
    """Define an igst object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Diagnostic(BaseModel):
    """Define a diagnostic object."""

    break_warn: BreakWarn = Field(..., alias="breakWarn")
    odos: Odos
    engine_oil_warn: EngineOilWarn = Field(..., alias="engineOilWarn")
    digsts: str
    tire_status: TireStatus = Field(..., alias="tireStatus")
    avail_range: AvailRange = Field(..., alias="availRange")
    odo: Odo
    spd: Spd
    igst: Igst


class Position(BaseModel):
    """Define a position object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class State(BaseModel):
    """Define a state object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Door(BaseModel):
    """Define a door object."""

    position: Position
    state: State


class DoorStatus(BaseModel):
    """Define a door status object."""

    id: str
    name: str
    doors: list[Door]


class State1(BaseModel):
    """Define a state1 object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Position1(BaseModel):
    """Define a position1 object."""

    name: str
    display_message: str = Field(..., alias="displayMessage")
    value: str


class Light(BaseModel):
    """Define a light object."""

    state: State1
    position: Position1


class LightStatus(BaseModel):
    """Define a light status object."""

    id: str
    name: str
    lights: list[Light]


class VehicleStatus(BaseModel):
    """Define a vehicle status object."""

    door_status: DoorStatus = Field(..., alias="doorStatus")
    light_status: LightStatus = Field(..., alias="lightStatus")


class Dt(BaseModel):
    """Define a dt object."""

    diagnostic: Diagnostic
    message: str
    vehicle_status: VehicleStatus = Field(..., alias="vehicleStatus")


class VhrItem(BaseModel):
    """Define a vhr item object."""

    cid: str
    vin: str
    operation: str
    ts: int
    dt: Dt


class VehicleStatusResponse(BaseModel):
    """Define a vehicle status response object."""

    vhr: list[VhrItem]



"""Define a token state object."""

from __future__ import annotations

from pydantic import BaseModel, Field


class TokenState(BaseModel):
    """Define a token state object."""

    access_token: str
    refresh_token: str
    refresh_expires_in: int
    token_type: str
    expires_in: int
    account_dn: str = Field(..., alias="accountDN")



"""Define a remote operation response object."""

from __future__ import annotations

from uuid import UUID  # noqa: TC003 needed for pydantic

from pydantic import BaseModel, Field


class RemoteOperationResponse(BaseModel):
    """Define a remote operation response object."""

    event_id: UUID = Field(..., alias="eventId")
    status_timestamp: str = Field(..., alias="statusTimestamp")
    start_time: str = Field(..., alias="startTime")
    operation_type: str = Field(..., alias="operationType")
    vin: str
    state: str
    status: str



"""Test the vehicle object."""

import json
import unittest
from datetime import date

from mitsubishi_connect_client.vehicle import VechiclesResponse

from . import sample_vehicle


class TestVehiclesResponse(unittest.TestCase):
    """Test the vehicles response object."""

    def test_parsing(self) -> None:
        """Test the from_text method."""
        loaded = json.loads(json.dumps(sample_vehicle))
        vehicles_response = VechiclesResponse(**loaded)
        self.assertEqual(len(vehicles_response.vehicles), 1)
        self.assertEqual(vehicles_response.vehicles[0].vin, "vin")
        vehicle = vehicles_response.vehicles[0]
        self.assertEqual(vehicle.vin, "vin")
        self.assertEqual(
            vehicle.date_of_sale,
            date(2024, 12, 29),
        )
        self.assertEqual(vehicle.primary_user, True)
        self.assertEqual(vehicle.make, "make")
        self.assertEqual(vehicle.model, "model")
        self.assertEqual(vehicle.year, 2024)
        self.assertEqual(vehicle.exterior_color_code, "exteriorColorCode")
        self.assertEqual(vehicle.exterior_color, "exteriorColor")
        self.assertEqual(vehicle.sim_state, "simState")
        self.assertEqual(vehicle.model_description, "modelDescription")
        self.assertEqual(vehicle.country, "country")
        self.assertEqual(vehicle.region, "region")
        self.assertEqual(vehicle.alpha_three_country_code, "alphaThreeCountryCode")
        self.assertEqual(vehicle.country_name, "countryName")
        self.assertEqual(vehicle.is_fleet, False)



"""Client for Mitsubishi Connect."""

import base64
import hashlib
import hmac
import json
import os
from typing import Any, ClassVar

import async_timeout
from aiohttp import ClientSession

from mitsubishi_connect_client.remote_operation_response import RemoteOperationResponse
from mitsubishi_connect_client.token_state import TokenState
from mitsubishi_connect_client.vehicle import VechiclesResponse
from mitsubishi_connect_client.vehicle_state import VehicleState
from mitsubishi_connect_client.vehicle_status import (
    VehicleStatusResponse,
    VhrItem,
)


class MitsubishiConnectClient:
    """Define the Mitsubishi Connect Client."""

    def __init__(
        self,
        user_name: str,
        password: str,
    ) -> None:
        """Initialize the client."""
        self._user_name = user_name
        self._password = password
        self._base_url = "https://us-m.aerpf.com"

    token: TokenState

    headers: ClassVar[dict[str, str]] = {
        "content-type": "application/json; charset=UTF-8",
        "user-agent": "Mobile",
        "x-client-id": "mobile",
        "ampapikey": "3f5547161b5d4bdbbb2bf8b26c69d1de",
        "host": "us-m.aerpf.com:15443",
        "connection": "Keep-Alive",
        "accept-encoding": "gzip",
    }

    async def login(self) -> None:
        """Login to the api."""
        url = f"{self._base_url}/auth/v1/token"
        data = {
            "grant_type": "password",
            "username": f"{self._user_name.strip()}",
            "password": f"{self._password.strip()}",
        }
        response = await self._api_wrapper(
            method="post",
            url=url,
            data=data,
            headers=self.headers,
        )
        self.token = TokenState(**response)

    async def refresh_token(self) -> None:
        """Login to the api."""
        url = f"{self._base_url}/auth/v1/token"
        data = {
            "grant_type": "refresh_token",
            "refresh_token": f"{self.token.refresh_token}",
        }
        response = await self._api_wrapper(
            method="post",
            url=url,
            data=data,
            headers=self.headers,
        )
        self.token = TokenState(**response)

    async def get_vehicles(self) -> VechiclesResponse:
        """Get the vehicles on the account."""
        url = f"{self._base_url}/user/v1/users/{self.token.account_dn}/vehicles"
        self.headers["authorization"] = "Bearer " + self.token.access_token
        response = await self._api_wrapper(
            method="get",
            url=url,
            headers=self.headers,
        )
        return VechiclesResponse(**response)

    async def get_vehicle_state(self, vin: str) -> VehicleState:
        """Get the vehicle state."""
        url = f"{self._base_url}/avi/v1/vehicles/{vin}/vehiclestate"
        self.headers["authorization"] = "Bearer " + self.token.access_token
        response = await self._api_wrapper(
            method="get",
            url=url,
            headers=self.headers,
        )
        return VehicleState(**response)

    def _get_bytes(self, data: dict[str, str]) -> bytes:
        """Get the bytes."""
        return json.dumps(data).replace(" ", "").encode("utf-8")

    async def _remote_operation_response(
        self, vin: str, operation: str
    ) -> RemoteOperationResponse:
        """Remote operation response."""
        url = f"{self._base_url}:15443/avi/v3/remoteOperation"
        data = {
            "vin": f"{vin}",
            "operation": f"{operation}",
            "forced": "true",
            "userAgent": "android",
        }
        headers = self._add_headers(self.headers, data)
        response = await self._api_wrapper(
            method="post",
            url=url,
            data_bytes=self._get_bytes(data),
            headers=headers,
        )
        return RemoteOperationResponse(**response)

    async def stop_engine(self, vin: str) -> RemoteOperationResponse:
        """Stop the engine."""
        operation = "engineOff"
        return await self._remote_operation_response(vin, operation)

    async def flash_lights(self, vin: str) -> RemoteOperationResponse:
        """Flash the lights."""
        operation = "lights"
        return await self._remote_operation_response(vin, operation)

    async def start_engine(self, vin: str) -> RemoteOperationResponse:
        """Start the engine."""
        operation = "remoteAC"
        return await self._remote_operation_response(vin, operation)

    async def unlock_vehicle(self, vin: str, pin_token: str) -> RemoteOperationResponse:
        """Unlock the vehicle."""
        url = f"{self._base_url}:15443/avi/v3/remoteOperation"
        data = {
            "vin": f"{vin}",
            "operation": "doorUnlock",
            "forced": "true",
            "pinToken": f"{pin_token}",
            "userAgent": "android",
        }
        headers = self._add_headers(self.headers, data)
        response = await self._api_wrapper(
            method="post",
            url=url,
            data_bytes=self._get_bytes(data),
            headers=headers,
        )
        return RemoteOperationResponse(**response)

    async def get_nonce(self, vin: str) -> dict[str, str]:
        """Get the server nonce."""
        url = f"{self._base_url}:15443/oauth/v3/remoteOperation"
        client_nonce = self._generate_client_nonce_base64()
        data = {
            "vin": f"{vin}",
            "clientNonce": f"{client_nonce}",
        }
        headers = self._add_headers(self.headers, data)
        response = await self._api_wrapper(
            method="post",
            url=url,
            data_bytes=self._get_bytes(data),
            headers=headers,
        )
        return {
            "clientNonce": client_nonce,
            "serverNonce": response["serverNonce"],
        }

    async def get_pin_token(self, vin: str, pin: str) -> str:
        """Get the pin token."""
        nonce = await self.get_nonce(vin)
        client_nonce = nonce["clientNonce"]
        server_nonce = nonce["serverNonce"]
        generated_hash = self._generate_hash(client_nonce, server_nonce, pin)
        url = f"{self._base_url}:15443/oauth/v3/remoteOperation/pin"
        client_nonce = self._generate_client_nonce_base64()
        data = {
            "vin": f"{vin}",
            "hash": f"{generated_hash}",
            "userAgent": "android",
        }
        headers = self._add_headers(self.headers, data)
        response = await self._api_wrapper(
            method="post",
            url=url,
            data_bytes=self._get_bytes(data),
            headers=headers,
        )
        return response["pinToken"]

    async def get_status(self, vin: str) -> VhrItem:
        """Get the vehicle status."""
        url = f"{self._base_url}:15443/avi/v1/vehicles/{vin}/vehicleStatus?count=1"
        self.headers["authorization"] = "Bearer " + self.token.access_token
        response = await self._api_wrapper(
            method="get",
            url=url,
            headers=self.headers,
        )
        return VehicleStatusResponse(**response).vhr[0]

    def _add_headers(
        self, headers: dict[str, str], data: dict[str, str]
    ) -> dict[str, str]:
        """Add headers to the request."""
        headers["authorization"] = "Bearer " + self.token.access_token
        data_bytes = self._get_bytes(data)
        length = len(data_bytes)
        headers["content-length"] = str(length)
        return headers

    async def _api_wrapper(
        self,
        method: str,
        url: str,
        headers: dict | None = None,
        data: Any | None = None,
        data_bytes: bytes | None = None,
    ) -> Any:
        """Get information from the API."""
        async with (
            async_timeout.timeout(10),
            ClientSession() as session,
            session.request(
                method=method,
                url=url,
                headers=headers,
                json=data,
                data=data_bytes,
            ) as response,
        ):
            if response.ok:
                text = await response.text()
                return json.loads(text)
            response.raise_for_status()
            return None

    def _generate_client_nonce_base64(self, length: int = 32) -> str:
        """Generate a random nonce and encodes it in Base64."""
        random_bytes = os.urandom(length)  # Generate random bytes
        return base64.b64encode(random_bytes).decode("utf-8")

    def _generate_hash(
        self, client_nonce: str, server_nonce: str, pin: str
    ) -> str | None:
        """Generate a custom hash based on client nonce, server nonce, and pin."""
        try:
            client_word_array = base64.b64decode(client_nonce)
            server_word_array = base64.b64decode(server_nonce)
            separator_word_array = b":"  # UTF-8 encoding

            # Construct the key (mimicking JavaScript concatenation)
            key_array = client_word_array + separator_word_array + server_word_array

            pin_array = pin.encode("utf-8")  # UTF-8 encoding

            hash256 = hmac.new(key_array, pin_array, hashlib.sha256).digest()

            hash128 = b""  # Initialize as bytes

            for i in range(4):
                word1 = hash256[i * 4 : (i * 4) + 4]  # Get 4 bytes
                word2 = hash256[(i + 4) * 4 : ((i + 4) * 4) + 4]
                word = bytes(x ^ y for x, y in zip(word1, word2, strict=False))
                # XOR the bytes
                hash128 += word

            return base64.b64encode(hash128).decode("utf-8")
        except (
            Exception  # noqa: BLE001
        ):
            return None



"""Define a vehicle state object."""

from __future__ import annotations

from datetime import datetime  # noqa: TC003 needed for pydantic

from pydantic import BaseModel, Field


class ExtLocMap(BaseModel):
    """Define an ext loc map object."""

    lon: float
    lat: float
    ts: datetime


class ChargingControl(BaseModel):
    """Define a charging control object."""

    cruising_range_combined: str = Field(..., alias="cruisingRangeCombined")
    event_timestamp: datetime = Field(..., alias="eventTimestamp")


class State(BaseModel):
    """Define a state object."""

    ext_loc_map: ExtLocMap = Field(..., alias="extLocMap")
    cst: bool
    tu_state: bool = Field(..., alias="tuState")
    ods: bool
    ignition_state: bool = Field(..., alias="ignitionState")
    odo: list[dict[datetime, int]]
    theft_alarm: bool = Field(..., alias="theftAlarm")
    svla: bool
    svtb: bool
    diagnostic: bool
    privacy: bool
    temp: str
    factory_reset: bool = Field(..., alias="factoryReset")
    tu_state_ts: datetime = Field(..., alias="tuStateTS")
    ignition_state_ts: datetime | None = Field(None, alias="ignitionStateTs")
    timezone: str
    accessible: bool
    charging_control: ChargingControl = Field(..., alias="chargingControl")


class VehicleState(BaseModel):
    """Define a vehicle state object."""

    vin: str
    ts: datetime
    state: State



"""Test the vehicle state object."""

import unittest
from datetime import UTC, datetime

from mitsubishi_connect_client.vehicle_state import VehicleState

from . import sample_vehicle_state


class TestVehicleState(unittest.TestCase):
    """Test the vehicle state object."""

    def test_vehiclestate_from_text(self) -> None:
        """Test VehicleState.from_text."""
        vehicle_state = VehicleState(**sample_vehicle_state)

        self.assertEqual(vehicle_state.vin, "1234567890ABCDEFG")
        self.assertEqual(
            vehicle_state.ts,
            datetime(2024, 3, 14, 12, 34, 56, 789000, tzinfo=UTC),
        )

        ext_loc_map = vehicle_state.state.ext_loc_map
        self.assertEqual(ext_loc_map.lon, 123.456)
        self.assertEqual(ext_loc_map.lat, 456.789)
        self.assertEqual(ext_loc_map.ts, datetime(2023, 3, 15, 13, 20, tzinfo=UTC))

        charging_control = vehicle_state.state.charging_control
        self.assertEqual(charging_control.cruising_range_combined, "200")
        self.assertEqual(
            charging_control.event_timestamp,
            datetime(2023, 3, 15, 13, 20, tzinfo=UTC),
        )
